#!/usr/bin/python3
import shutil, os, sys

# This scripts installs the PlasmidFinder database for using KMA
# KMA should be installed before running this script
# The scripts is assumed to be located together with plasmidfinder's '.fsa' files
#
# First clone the repository: git clone https://bitbucket.org/genomicepidemiology/plasmidfinder_db.git

# Check if executable kma_index program is installed, if not promt the user for path

interactive = True
if len(sys.argv) >= 2:
    kma_index = sys.argv[1]
    if "non_interactive" in sys.argv:
        interactive = False
else:
    kma_index = "kma_index"

while shutil.which(kma_index) is None:
    if not interactive:
        sys.exit("KMA index program, {}, does not exist or is not executable".format(kma_index))
    ans = input("Please input path to executable kma_index program or enter 'q'/'quit' to exit:")
    if ans == "q" or ans == "quit":
        print("Exiting!\n\n \
               Please install executable KMA programs in order to install this database.\n\n \
               KMA can be obtained from bitbucked:\n\n\
               git clone https://bitbucket.org/genomicepidemiology/kma.git\n\n\
               KMA programs must afterwards be compiled:\n\n\
               gcc -O3 -o kma KMA.c -lm -lpthread\n\
               gcc -O3 -o kma_index KMA_index.c -lm")
        sys.exit()

    kma_index = ans

    if shutil.which(kma_index) is None:
        print("Path, {}, is not an executable path. Please provide absolute path\n".format(ans))


# Index databases

# Use config_file to go through database dirs
config_file = open("config", "r")
for line in config_file:
    if line.startswith("#"):
        continue
    else:
        line = line.rstrip().split("\t")
        species = line[0] 
        # for each dir index the fasta files
        os.system("{0} -i {1}.fsa -o {1}".format(kma_index, species))
config_file.close() 

print("Done")



from collections import OrderedDict

from staramr.results.AMRDetectionSummary import AMRDetectionSummary

"""
Summarizes both ResFinder and PointFinder database results into a single table.
"""


class AMRDetectionSummaryResistance(AMRDetectionSummary):

    def __init__(self, files, resfinder_dataframe, pointfinder_dataframe=None, plasmidfinder_dataframe=None):
        """
        Creates a new AMRDetectionSummaryResistance.
        :param files: The list of genome files we have scanned against.
        :param resfinder_dataframe: The pd.DataFrame containing the ResFinder results.
        :param pointfinder_dataframe: The pd.DataFrame containing the PointFinder results.
        :param plasmidfinder_dataframe: The pd.DataFrame containing the PlasmidFinder results.
        """
        super().__init__(files, resfinder_dataframe, pointfinder_dataframe, plasmidfinder_dataframe)

    def _aggregate_gene_phenotype(self, dataframe):
        flattened_phenotype_list = [y.strip() for x in dataframe['Predicted Phenotype'].tolist() for y in
                                    x.split(self.SEPARATOR)]
        uniq_phenotype = OrderedDict.fromkeys(flattened_phenotype_list)

        return {'Gene': (self.SEPARATOR + ' ').join(dataframe['Gene']),
                'Predicted Phenotype': (self.SEPARATOR + ' ').join(list(uniq_phenotype))
                }

    def _compile_results(self, df):
        df_summary = df.copy()

        # Used to sort by gene names, ignoring case
        df_summary['Gene.Lower'] = df['Gene'].str.lower()

        # Compiles the gene/phenotype results into a single entry per isolate (groupby)
        df_summary = df_summary \
            .sort_values(by=['Gene.Lower']) \
            .groupby(['Isolate ID'], sort=True) \
            .aggregate(self._aggregate_gene_phenotype)
        return df_summary[['Gene', 'Predicted Phenotype']]

    def _get_detailed_negative_columns(self):
        return ['Isolate ID', 'Gene', 'Predicted Phenotype', 'Start', 'End']

    def _get_summary_empty_values(self):
        return {'Genotype': 'None', 'Predicted Phenotype': 'Sensitive'}

    def _get_summary_resistance_columns(self):
        return ['Genotype', 'Predicted Phenotype', 'Plasmid Genes']

    def _get_detailed_summary_columns(self):
        return ['Gene', 'Predicted Phenotype', '%Identity', '%Overlap', 'HSP Length/Total Length', 'Contig', 'Start',
                'End', 'Accession', 'Data Type']

    def _include_phenotype(self):
        return True



import logging
import re
from typing import List

import pandas as pd

from staramr.blast.results.AMRHitHSP import AMRHitHSP

logger = logging.getLogger('PlasmidfinderHitHSP')

"""
A Class storing a PlasmidFinder-specific BLAST hit/HSP.
"""


class PlasmidfinderHitHSP(AMRHitHSP):

    def __init__(self, file: str, blast_record: pd.Series) -> None:
        """
        Builds a new PlasmidfinderHitHSP.
        :param file: The input file.
        :param blast_record: The Bio.Blast.Record this hit came from.
        :param hit: The particular Bio.Blast.Record.Alignment.
        :param hsp: The particular Bio.Blast.Record.HSP.
        """
        super().__init__(file, blast_record)

        logger.debug("record=%s", self._blast_record)

        splitList = re.split('_', self.get_amr_gene_id())
        re_search = list(filter(None, splitList))  # type: List[str]

        if not re_search:
            raise Exception("Could not split up seq name for [" + self.get_amr_gene_id() + "]")

        length = len(re_search)

        # Add empty string if gene variant is missing
        if length == 2:
            re_search.insert(1, '')
            length += 1

        if 3 <= length <= 5:
            self._gene = re_search[0]
            self._gene_variant = re_search[1]

            if length == 3:
                self._accession = re_search[2]
            elif length == 4:
                self._accession = re_search[3]
            elif length == 5:
                self._accession = re_search[3] + "_" + re_search[4]
            else:
                raise Exception("Length not valid for variable assignment, for [" + self.get_amr_gene_id() + "]")
        else:
            raise Exception("Index length not within range, for [" + self.get_amr_gene_id() + "]")

    def get_amr_gene_name(self) -> str:
        """
        Gets the gene name for the PlasmidFinder hit.
        :return: The gene name.
        """
        return self._gene

    def get_amr_gene_name_with_variant(self) -> str:
        """
        Gets the gene name + variant number for the PlasmidFinder hit.
        :return: The gene name + variant number.
        """
        return self.get_amr_gene_name() + '_' + self._gene_variant

    def get_amr_gene_variant_accession(self) -> str:
        """
        Gets the gene name + variant number + accession for the PlasmidFinder hit.
        :return: The gene name + variant number + accession.
        """
        return self._gene + '_' + self._gene_variant + '_' + self._accession

    def get_amr_gene_variant(self) -> str:
        """
        Gets the variant number for the PlasmidFinder hit.
        :return: The variant number.
        """
        return self._gene_variant

    def get_amr_gene_accession(self) -> str:
        """
        Gets the accession for the PlasmidFinder hit.
        :return: The accession.
        """
        return self._accession



import logging
from os import path
from typing import Set

import pandas as pd
from pandas import DataFrame

logger = logging.getLogger("AMRDetectionSummary")

"""
Summarizes both ResFinder, PointFinder, and PlasmidFinder database results into a single table.
"""


class AMRDetectionSummary:
    SEPARATOR = ','
    FLOAT_DECIMALS = 2

    def __init__(self, files, resfinder_dataframe: DataFrame, pointfinder_dataframe=None,
                 plasmidfinder_dataframe=None) -> None:
        """
        Constructs an object for summarizing AMR detection results.
        :param files: The list of genome files we have scanned against.
        :param resfinder_dataframe: The pd.DataFrame containing the ResFinder results.
        :param pointfinder_dataframe: The pd.DataFrame containing the PointFinder results.
        """
        self._names = [path.splitext(path.basename(x))[0] for x in files]
        self._resfinder_dataframe = resfinder_dataframe
        self._plasmidfinder_dataframe = plasmidfinder_dataframe

        if pointfinder_dataframe is not None:
            self._has_pointfinder = True
            self._pointfinder_dataframe = pointfinder_dataframe
        else:
            self._has_pointfinder = False

    def _compile_results(self, resistance_frame: DataFrame) -> DataFrame:
        df_summary = resistance_frame.sort_values(by=['Gene']).groupby(['Isolate ID']).aggregate(
            lambda x: {'Gene': (self.SEPARATOR + ' ').join(x['Gene'])})
        return df_summary[['Gene']]

    def _compile_plasmids(self, plasmid_frame: DataFrame) -> DataFrame:
        ds_summary = plasmid_frame.sort_values(by=['Gene']).groupby(['Isolate ID']).aggregate(
            lambda x: {'Gene': (self.SEPARATOR + ' ').join(x['Gene'])})

        ds_frame = ds_summary[['Gene']]

        plasmid_frame = ds_frame.rename(columns={'Gene': 'Plasmid Genes'})

        return plasmid_frame

    def _include_negatives(self, resistance_frame: DataFrame) -> DataFrame:
        result_names_set = set(resistance_frame.index.tolist())
        names_set = set(self._names)

        negative_names_set = names_set - result_names_set
        negative_entries = pd.DataFrame([[x, 'None'] for x in negative_names_set],
                                        columns=('Isolate ID', 'Gene')).set_index('Isolate ID')

        return resistance_frame.append(negative_entries, sort=True)

    def _get_detailed_negative_columns(self):
        return ['Isolate ID', 'Gene', 'Start', 'End']

    def _include_phenotype(self):
        return False

    def _get_negative_resistance_entries(self, names_set: Set, resistance_frame: DataFrame) -> DataFrame:
        resfinder_names_set = set(resistance_frame.index.tolist())
        negative_res_names_set = names_set - resfinder_names_set
        negative_columns = self._get_detailed_negative_columns()

        negative_resistance_entries = None

        if len(negative_res_names_set) != len(names_set) or resistance_frame.empty:

            if self._include_phenotype():
                negative_resistance_entries = pd.DataFrame(
                    [[x, 'None', 'Sensitive', '', ''] for x in negative_res_names_set],
                    columns=negative_columns).set_index('Isolate ID')
            else:
                negative_resistance_entries = pd.DataFrame([[x, 'None', '', ''] for x in negative_res_names_set],
                                                           columns=negative_columns).set_index('Isolate ID')

            negative_resistance_entries['Data Type'] = 'Resistance'

        return negative_resistance_entries

    def _include_detailed_negatives(self, resistance_frame: DataFrame, plasmid_frame: DataFrame = None) -> DataFrame:
        names_set = set(self._names)
        set_used = names_set

        negative_entries = self._get_negative_resistance_entries(names_set, resistance_frame)

        if plasmid_frame is not None:
            plasmid_frame = self._compile_plasmids(plasmid_frame)
            plasmidfinder_names_set = set(plasmid_frame.index.tolist())
            negative_plasmid_names_set = names_set - plasmidfinder_names_set

            if not plasmid_frame.empty:
                set_used = negative_plasmid_names_set

        negative_plasmid_entries = pd.DataFrame([[x, 'None'] for x in set_used],
                                                columns=('Isolate ID', 'Gene')).set_index('Isolate ID')
        negative_plasmid_entries['Data Type'] = 'Plasmid'

        if negative_entries is None:
            negative_entries = negative_plasmid_entries
        else:
            negative_entries = negative_entries.append(negative_plasmid_entries, sort=True)

        return resistance_frame.append(negative_entries, sort=True)

    def _get_summary_empty_values(self):
        return {'Genotype': 'None'}

    def _get_summary_resistance_columns(self):
        return ['Genotype', 'Plasmid Genes']

    def create_summary(self, include_negatives: bool = False) -> DataFrame:
        """
        Constructs a summary pd.DataFrame for all ResFinder/PointFinder/PlasmidFinder results.
        :param include_negatives: If True, include files with no ResFinder/PointFinder/PlasmidFinder results.
        :return: A pd.DataFrame summarizing the results.
        """
        resistance_frame = self._resfinder_dataframe
        plasmid_frame = self._plasmidfinder_dataframe

        if self._has_pointfinder:
            resistance_frame = resistance_frame.append(self._pointfinder_dataframe, sort=True)

        resistance_frame = self._compile_results(resistance_frame)

        if include_negatives:
            resistance_frame = self._include_negatives(resistance_frame)

        resistance_frame.rename(columns={'Gene': 'Genotype'}, inplace=True)

        fill_values = self._get_summary_empty_values()
        resistance_columns = self._get_summary_resistance_columns()

        if plasmid_frame is not None:
            plasmid_frame = self._compile_plasmids(plasmid_frame)

            if resistance_frame.empty:
                resistance_frame = resistance_frame.append(plasmid_frame)
            else:
                resistance_frame = resistance_frame.merge(plasmid_frame, on='Isolate ID', how='left').fillna(
                    value={'Plasmid Genes': 'None'})

            resistance_frame = resistance_frame.fillna(value=fill_values)
            resistance_frame = resistance_frame.reindex(columns=resistance_columns)

        return resistance_frame.sort_index()

    def _get_detailed_summary_columns(self):
        return ['Gene', '%Identity', '%Overlap', 'HSP Length/Total Length', 'Contig', 'Start', 'End', 'Accession',
                'Data Type']

    def create_detailed_summary(self, include_negatives: bool = True) -> DataFrame:
        if self._resfinder_dataframe is None:
            resistance_frame = None
        else:
            resistance_frame = self._resfinder_dataframe.copy()
            resistance_frame['Data Type'] = 'Resistance'
            resistance_frame = resistance_frame.round(
                {'%Identity': self.FLOAT_DECIMALS, '%Overlap': self.FLOAT_DECIMALS})

        if self._plasmidfinder_dataframe is None:
            plasmid_frame = None
        else:
            plasmid_frame = self._plasmidfinder_dataframe.copy()

        column_names = self._get_detailed_summary_columns()

        if self._has_pointfinder:
            if self._pointfinder_dataframe is None:
                point_frame = None
            else:
                point_frame = self._pointfinder_dataframe.copy()
                point_frame['Data Type'] = 'Resistance'
                point_frame = point_frame.round({'%Identity': self.FLOAT_DECIMALS, '%Overlap': self.FLOAT_DECIMALS})
                point_frame = point_frame.reindex(columns=column_names)

            if resistance_frame is not None:
                resistance_frame = resistance_frame.append(point_frame, sort=True)

        if include_negatives:
            if plasmid_frame is not None:
                plasmid_frame = plasmid_frame.reindex(columns=column_names)
                resistance_frame = self._include_detailed_negatives(resistance_frame, plasmid_frame)
            else:
                resistance_frame = self._include_detailed_negatives(resistance_frame)
            resistance_frame = resistance_frame.reindex(columns=column_names)

        if plasmid_frame is not None:
            plasmid_frame['Data Type'] = 'Plasmid'

            if self._include_phenotype():
                plasmid_frame['Predicted Phenotype'] = ''

            plasmid_frame = plasmid_frame.round({'%Identity': self.FLOAT_DECIMALS, '%Overlap': self.FLOAT_DECIMALS})

            if resistance_frame is not None:
                resistance_frame = resistance_frame.append(plasmid_frame, sort=True)
                resistance_frame = resistance_frame.reindex(columns=column_names)
                resistance_frame = resistance_frame.sort_values(['Isolate ID', 'Data Type', 'Gene'])

        if resistance_frame is not None:
            resistance_frame = resistance_frame.fillna("")

        return resistance_frame



import logging
import os
import re
import subprocess
from concurrent.futures import ThreadPoolExecutor
from os import path
from typing import Dict

from Bio.Blast.Applications import NcbiblastnCommandline

from staramr.blast.AbstractBlastDatabase import AbstractBlastDatabase
from staramr.exceptions.BlastProcessError import BlastProcessError

logger = logging.getLogger('BlastHandler')

"""
Class for handling scheduling of BLAST jobs.
"""


class BlastHandler:
    BLAST_COLUMNS = [x.strip() for x in '''
    qseqid
    sseqid
    pident
    length
    qstart
    qend
    sstart
    send
    slen
    qlen
    sstrand
    sseq
    qseq
    '''.strip().split('\n')]

    def __init__(self, blast_database_objects_map: Dict[str, AbstractBlastDatabase], threads: int,
                 output_directory: str) -> None:
        """
        Creates a new BlastHandler.
        :param blast_database_objects_map: A map containing the blast databases.
        :param threads: The maximum number of threads to use, where one BLAST process gets assigned to one thread.
        :param output_directory: The output directory to store BLAST results.
        """
        if threads is None:
            raise Exception("threads is None")

        self._threads = threads

        if output_directory is None:
            raise Exception("output_directory is None")

        self._output_directory = output_directory
        self._input_genomes_tmp_dir = path.join(output_directory, 'input-genomes')

        self._blast_database_objects_map = blast_database_objects_map

        if (self._blast_database_objects_map['pointfinder'] is None):
            self._pointfinder_configured = False  # type: bool
            del self._blast_database_objects_map['pointfinder']
        else:
            self._pointfinder_configured = True  # type: bool

        self._thread_pool_executor = None
        self.reset()

    def reset(self):
        """
        Resets this BlastHandler.
        :return: None
        """
        if self._thread_pool_executor is not None:
            self._thread_pool_executor.shutdown()
        self._thread_pool_executor = ThreadPoolExecutor(max_workers=self._threads)
        self._blast_map = {}
        self._future_blasts_map = {}

        if path.exists(self._input_genomes_tmp_dir):
            logger.debug("Directory [%s] already exists", self._input_genomes_tmp_dir)
        else:
            os.mkdir(self._input_genomes_tmp_dir)

    def run_blasts(self, files) -> None:
        """
        Scans all files with BLAST against the ResFinder/PointFinder/Plasmid databases.
        :param files: The files to scan.
        :return: None
        """
        db_files = self._make_db_from_input_files(self._input_genomes_tmp_dir, files)
        logger.debug("Done making blast databases for input files")

        for file in db_files:
            logger.info("Scheduling blasts for %s", path.basename(file))

            for name in self._blast_database_objects_map:
                database_object = self._blast_database_objects_map[name]
                self._schedule_blast(file, database_object)

    def _make_db_from_input_files(self, db_dir, files):
        logger.info("Making BLAST databases for input files")
        future_makeblastdbs = []
        db_files = []

        for file in files:
            destination = path.join(db_dir, path.basename(file))
            logger.debug("Creating symlink from [%s] to [%s]", file, destination)
            os.symlink(path.abspath(file), destination)
            db_files.append(destination)

            future_makeblastdbs.append(self._thread_pool_executor.submit(self._make_blast_db, destination, file))

        # Blocks until all blast dbs are made. If an exception is raised, will raise same exception
        try:
            for future_blastdb in future_makeblastdbs:
                future_blastdb.result()
        except subprocess.CalledProcessError as e:
            raise BlastProcessError("Error running makeblastdb", e)

        return db_files

    def _schedule_blast(self, file, blast_database):
        database_names = blast_database.get_database_names()
        logger.debug("%s databases: %s", blast_database.get_name(), database_names)
        for database_name in database_names:
            database = blast_database.get_path(database_name)
            file_name = os.path.basename(file)

            blast_out = os.path.join(self._output_directory,
                                     file_name + "." + database_name + "." + blast_database.get_name() + ".blast.tsv")
            if os.path.exists(blast_out):
                raise Exception("Error, blast_out [%s] already exists", blast_out)

            self._get_blast_map(blast_database.get_name()).setdefault(file_name, {})[database_name] = blast_out

            future_blast = self._thread_pool_executor.submit(self._launch_blast, database, file, blast_out)
            self._get_future_blasts_from_map(blast_database.get_name()).append(future_blast)

    def _get_blast_map(self, name: str) -> Dict:
        if name not in self._blast_map:
            self._blast_map[name] = {}

        return self._blast_map[name]

    def _get_future_blasts_from_map(self, name: str) -> Dict:
        if name not in self._future_blasts_map:
            self._future_blasts_map[name] = []

        return self._future_blasts_map[name]

    def is_pointfinder_configured(self) -> bool:
        """
        Whether or not PointFinder is being used.
        :return: True if PointFinder is being used, False otherwise.
        """
        return self._pointfinder_configured

    def get_resfinder_outputs(self) -> Dict:
        """
        Gets the ResFinder output files in the form of a dictionary which looks like:
            { 'input_file_name' => 'blast_results_file.xml' }
        :return: A dictionary mapping input file names to ResFinder BLAST output files.
        """

        # Forces any exceptions to be thrown if error with blasts
        for future_blast in self._get_future_blasts_from_map('resfinder'):
            future_blast.result()
        return self._get_blast_map('resfinder')

    def get_plasmidfinder_outputs(self) -> Dict:
        """
        Gets the PlasmidFinder output files in the form of a dictionary which looks like:
            { 'input_file_name' => 'blast_results_file.xml' }
        :return: A dictionary mapping input file names to ResFinder BLAST output files.
        """

        # Forces any exceptions to be thrown if error with blasts
        for future_blast in self._get_future_blasts_from_map('plasmidfinder'):
            future_blast.result()
        return self._get_blast_map('plasmidfinder')

    def get_pointfinder_outputs(self) -> Dict:
        """
        Gets the PointFinder output files in the form of a dictionary which looks like:
            { 'input_file_name' => 'blast_results_file.xml' }
        :return: A dictionary mapping input file names to PointFinder BLAST output files.
        """
        if (self.is_pointfinder_configured()):
            # Forces any exceptions to be thrown if error with blasts
            for future_blast in self._get_future_blasts_from_map('pointfinder'):
                future_blast.result()
            return self._get_blast_map('pointfinder')
        else:
            raise Exception('Error, pointfinder has not been configured')

    def _launch_blast(self, query, db, output) -> None:
        blast_out_format = '"6 ' + ' '.join(self.BLAST_COLUMNS) + '"'

        blastn_command = NcbiblastnCommandline(query=query, db=db, evalue=0.001, outfmt=blast_out_format, out=output)
        stdout, stderr = blastn_command()

        if stderr:
            raise Exception("error with [" + str(blastn_command) + "], stderr=" + stderr)

    def _make_blast_db(self, path: str, file: str) -> None:
        command = ['makeblastdb', '-in', path, '-dbtype', 'nucl', '-parse_seqids']
        logger.debug(' '.join(command))
        try:
            subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        except subprocess.CalledProcessError as e:
            err_msg = str(e.stderr.strip())
            err_msg = re.findall('REF\|(.*?)\'', err_msg)[0]
            raise Exception('Could not run makeblastdb on file {}, error {}'.format(file, err_msg))



from os import path

import pandas as pd

"""
A Class used to parse out a list of genes to exclude from the results.
"""


class ExcludeGenesList:
    DEFAULT_EXCLUDE_FILE = path.join(path.dirname(__file__), 'data', 'genes_to_exclude.tsv')

    def __init__(self, file=DEFAULT_EXCLUDE_FILE):
        self._data = pd.read_csv(file, sep='\t')

    def tolist(self):
        """
        Converts the exclude genes data to a list.
        :return: A list with genes to exclude.
        """
        return self._data['#gene_id'].tolist()

    @classmethod
    def get_default_exclude_file(cls):
        """
        Get the default file containing the list of genes to exclude.
        :return: The default file containing the list of genes to exclude.
        """
        return cls.DEFAULT_EXCLUDE_FILE



import logging
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

logger = logging.getLogger('BlastHits')

from staramr.blast.results.AMRHitHSP import AMRHitHSP
from staramr.exceptions.InvalidPositionException import InvalidPositionException

"""
Class for partitioning up blast hits into non-overlapping regions.
"""


class BlastHitPartitions:

    def __init__(self):
        """
        Creates a new object to store BLAST hit partitions.
        """
        self._partitions = OrderedDict()

    def append(self, hit: AMRHitHSP) -> None:
        """
        Adds a new blast hit to the set of partitions.
        :param hit: The hit to add.
        :return: None
        """
        if hit.get_genome_contig_start() > hit.get_genome_contig_end() and hit.get_genome_contig_strand() == 'plus':
            raise InvalidPositionException(
                "Unsupported condition: strand=plus and contig start > contig end for hit (contig=" + hit.get_genome_contig_id() + ", start=" +
                str(hit.get_genome_contig_start()) + ", end=" + str(hit.get_genome_contig_end()) + ")")

        partition = self._get_existing_partition(hit)
        if (partition is None):
            self._create_new_parition(hit)
        else:
            self._add_hit_partition(hit, partition)

    def _add_hit_partition(self, hit: AMRHitHSP, partition: Dict[str, Any]) -> None:
        start, end = self._stranded_ends(hit)

        if start < partition['start']:
            partition['start'] = start

        if end > partition['end']:
            partition['end'] = end

        partition['hits'].append(hit)

    def _get_existing_partition(self, hit: AMRHitHSP) -> Optional[Dict[str, Any]]:
        partition_name = hit.get_genome_contig_id()

        if partition_name in self._partitions:
            contig_partitions_list = self._partitions[partition_name]
            for partition in contig_partitions_list:
                if self._hit_in_parition(hit, partition):
                    return partition

        return None

    def _hit_in_parition(self, hit: AMRHitHSP, partition: Dict[str, Any]) -> bool:
        pstart, pend = partition['start'], partition['end']
        start, end = self._stranded_ends(hit)

        return (pstart < start < pend) or (pstart < end < pend) or (start <= pstart and end >= pend)

    def _create_new_parition(self, hit: AMRHitHSP) -> None:
        start, end = self._stranded_ends(hit)

        contig_name = hit.get_genome_contig_id()
        partition = {
            'start': start,
            'end': end,
            'hits': [hit]
        }

        if contig_name in self._partitions:
            self._partitions[contig_name].append(partition)
        else:
            self._partitions[contig_name] = [partition]

    def get_hits_nonoverlapping_regions(self) -> List[List[AMRHitHSP]]:
        """
        Gets BLAST hits divided up into separate lists for non-overlapping regions..
        :return: A list of BLAST hits divided up into non-overlapping regions.
        """
        return [p['hits'] for name in self._partitions for p in self._partitions[name]]

    def _stranded_ends(self, hit: AMRHitHSP) -> Tuple[int, int]:
        """
        Gets the start/end coordinates, taking into account the strand.
        :param hit: The hit.
        :return: The (start,end) as a tuple.
        """
        start = hit.get_genome_contig_start() if hit.get_genome_contig_strand() == 'plus' else hit.get_genome_contig_end()
        end = hit.get_genome_contig_end() if hit.get_genome_contig_strand() == 'plus' else hit.get_genome_contig_start()
        return start, end



"""
Classes for interacting with the (ResFinder/PointFinder/PlasmidFinder) databases used to detect AMR genes.
"""
import argparse
import logging
import sys
from os import path, mkdir

from staramr.SubCommand import SubCommand
from staramr.Utils import get_string_with_spacing
from staramr.databases.AMRDatabasesManager import AMRDatabasesManager
from staramr.databases.resistance.ARGDrugTable import ARGDrugTable
from staramr.exceptions.CommandParseException import CommandParseException
from staramr.exceptions.DatabaseErrorException import DatabaseErrorException
from staramr.exceptions.DatabaseNotFoundException import DatabaseNotFoundException

"""
Base class for interacting with a database.
"""

logger = logging.getLogger('AMRDatabaseHandler')


class Database(SubCommand):

    def __init__(self, subparser, script_name):
        """
        Builds a SubCommand for interacting with databases.
        :param subparser: The subparser to use.  Generated from argparse.ArgumentParser.add_subparsers().
        :param script_name: The name of the script being run.
        """
        super().__init__(subparser, script_name)

    def _setup_args(self, arg_parser):
        arg_parser = self._subparser.add_parser('db', help='Download ResFinder/PointFinder/PlasmidFinder databases')
        subparser = arg_parser.add_subparsers(dest='db_command',
                                              help='Subcommand for ResFinder/PointFinder/PlasmidFinder databases.')

        Build(subparser, self._script_name + " db")
        Update(subparser, self._script_name + " db")
        Info(subparser, self._script_name + " db")
        RestoreDefault(subparser, self._script_name + " db")

        return arg_parser

    def run(self, args):
        super(Database, self).run(args)

        if args.db_command is None:
            self._root_arg_parser.print_help()


"""
Class for building a new database.
"""


class Build(Database):

    def __init__(self, subparser, script_name):
        """
        Creates a SubCommand for building a new database.
        :param subparser: The subparser to use.  Generated from argparse.ArgumentParser.add_subparsers().
        :param script_name: The name of the script being run.
        """
        super().__init__(subparser, script_name)

    def _setup_args(self, arg_parser):
        name = self._script_name
        self._default_dir = AMRDatabasesManager.get_default_database_directory()
        epilog = ("Example:\n"
                  "\t" + name + " build\n"
                                "\t\tBuilds a new ResFinder/PointFinder/PlasmidFinder database under " + self._default_dir + " if it does not exist\n\n" +
                  "\t" + name + " build --dir databases\n" +
                  "\t\tBuilds a new ResFinder/PointFinder/PlasmidFinder database under databases/")

        arg_parser = self._subparser.add_parser('build',
                                                epilog=epilog,
                                                formatter_class=argparse.RawTextHelpFormatter,
                                                help='Downloads and builds databases in the given directory.')
        arg_parser.add_argument('--dir', action='store', dest='destination', type=str,
                                help='The directory to download the databases into [' + self._default_dir + '].',
                                default=self._default_dir, required=False)
        arg_parser.add_argument('--resfinder-commit', action='store', dest='resfinder_commit', type=str,
                                help='The specific git commit for the resfinder database [latest].', required=False)
        arg_parser.add_argument('--pointfinder-commit', action='store', dest='pointfinder_commit', type=str,
                                help='The specific git commit for the pointfinder database [latest].', required=False)
        arg_parser.add_argument('--plasmidfinder-commit', action='store', dest='plasmidfinder_commit', type=str,
                                help='The specific git commit for the plasmidfinder database [latest].', required=False)
        return arg_parser

    def run(self, args):
        super(Build, self).run(args)

        if path.exists(args.destination):
            if args.destination == self._default_dir:
                raise CommandParseException("Error, default destination [" + args.destination + "] already exists",
                                            self._root_arg_parser, print_help=True)
            else:
                raise CommandParseException("Error, destination [" + args.destination + "] already exists",
                                            self._root_arg_parser)
        else:
            mkdir(args.destination)

        if args.destination == AMRDatabasesManager.get_default_database_directory():
            database_repos = AMRDatabasesManager.create_default_manager().get_database_repos()
        else:
            database_repos = AMRDatabasesManager(args.destination).get_database_repos()
        database_repos.build({'resfinder': args.resfinder_commit, 'pointfinder': args.pointfinder_commit,
                              'plasmidfinder': args.plasmidfinder_commit})
        if not AMRDatabasesManager.is_database_repos_default_commits(database_repos):
            logger.warning(
                "Built non-default ResFinder/PointFinder/PlasmidFinder database version. This may lead to " +
                "differences in the detected AMR genes depending on how the database files are structured.")


"""
Class for updating an existing database.
"""


class Update(Database):

    def __init__(self, subparser, script_name):
        """
        Creates a SubCommand for updating an existing database.
        :param subparser: The subparser to use.  Generated from argparse.ArgumentParser.add_subparsers().
        :param script_name: The name of the script being run.
        """
        super().__init__(subparser, script_name)

    def _setup_args(self, arg_parser):
        self._default_dir = AMRDatabasesManager.get_default_database_directory()
        name = self._script_name
        epilog = ("Example:\n"
                  "\t" + name + " update databases/\n"
                                "\t\tUpdates the ResFinder/PointFinder/PlasmidFinder database under databases/\n\n" +
                  "\t" + name + " update -d\n" +
                  "\t\tUpdates the default ResFinder/PointFinder/PlasmidFinder database under " + self._default_dir)
        arg_parser = self._subparser.add_parser('update',
                                                epilog=epilog,
                                                formatter_class=argparse.RawTextHelpFormatter,
                                                help='Updates databases in the given directories.')

        arg_parser.add_argument('-d', '--update-default', action='store_true', dest='update_default',
                                help='Updates default database directory (' + self._default_dir + ').', required=False)
        arg_parser.add_argument('--resfinder-commit', action='store', dest='resfinder_commit', type=str,
                                help='The specific git commit for the resfinder database [latest].', required=False)
        arg_parser.add_argument('--pointfinder-commit', action='store', dest='pointfinder_commit', type=str,
                                help='The specific git commit for the pointfinder database [latest].', required=False)
        arg_parser.add_argument('--plasmidfinder-commit', action='store', dest='plasmidfinder_commit', type=str,
                                help='The specific git commit for the plasmidfinder database [latest].', required=False)
        arg_parser.add_argument('directories', nargs='*')

        return arg_parser

    def run(self, args):
        super(Update, self).run(args)

        if len(args.directories) == 0:
            if not args.update_default:
                raise CommandParseException("Must pass at least one directory to update, or use '--update-default'",
                                            self._root_arg_parser,
                                            print_help=True)
            else:
                try:
                    database_repos = AMRDatabasesManager.create_default_manager().get_database_repos(
                        force_use_git=True)

                    database_repos.update(
                        {'resfinder': args.resfinder_commit, 'pointfinder': args.pointfinder_commit,
                         'plasmidfinder': args.plasmidfinder_commit})

                    if not AMRDatabasesManager.is_database_repos_default_commits(database_repos):
                        logger.warning(
                            "Updated to non-default ResFinder/PointFinder/PlasmidFinder database version. This may lead to " +
                            "differences in the detected AMR genes depending on how the database files are structured.")
                except DatabaseErrorException as e:
                    logger.error(
                        "Could not update default database. Please try restoring with 'staramr db restore-default'")
                    raise e
        else:
            for directory in args.directories:
                database_repos = AMRDatabasesManager(directory).get_database_repos()
                database_repos.update({'resfinder': args.resfinder_commit, 'pointfinder': args.pointfinder_commit,
                                       'plasmidfinder': args.plasmidfinder_commit})
                if not AMRDatabasesManager.is_database_repos_default_commits(database_repos):
                    logger.warning(
                        "Updated to non-default ResFinder/PointFinder/PlasmidFinder database version [%s]. This may lead to " +
                        "differences in the detected AMR genes depending on how the database files are structured.",
                        directory)


"""
Class for restoring the default database.
"""


class RestoreDefault(Database):

    def __init__(self, subparser, script_name):
        """
        Creates a SubCommand for restoring to the default database.
        :param subparser: The subparser to use.  Generated from argparse.ArgumentParser.add_subparsers().
        :param script_name: The name of the script being run.
        """
        super().__init__(subparser, script_name)

    def _setup_args(self, arg_parser):
        name = self._script_name
        epilog = ("Example:\n"
                  "\t" + name + " restore-default\n"
                                "\t\tRestores the default ResFinder/PointFinder/PlasmidFinder database\n\n")
        arg_parser = self._subparser.add_parser('restore-default',
                                                epilog=epilog,
                                                formatter_class=argparse.RawTextHelpFormatter,
                                                help='Restores the default ResFinder/PointFinder/PlasmidFinder databases.')

        arg_parser.add_argument('-f', '--force', action='store_true', dest='force',
                                help='Force restore without asking for confirmation.', required=False)
        return arg_parser

    def _confirm_restore(self):
        """
        Confirms with the user whether or not to restore the database directory.
        :return: True if should restore, False otherwise.
        """
        confirmed = False
        while not confirmed:
            response = str(input(
                "Restore the default ResFinder/PointFinder/PlasmidFinder databases (Y/N)? ").lower().strip())
            if response == 'y' or response == 'yes':
                return True
            elif response == 'n' or response == 'no':
                return False

    def run(self, args):
        super(RestoreDefault, self).run(args)

        database_manager = AMRDatabasesManager.create_default_manager()

        if not args.force:
            response = self._confirm_restore()
        else:
            response = True

        if response:
            database_manager.restore_default()


"""
Class for getting information from an existing database.
"""


class Info(Database):

    def __init__(self, subparser, script_name):
        """
        Creates a SubCommand for printing information about a database.
        :param subparser: The subparser to use.  Generated from argparse.ArgumentParser.add_subparsers().
        :param script_dir: The directory containing the main application script.
        :param script_name: The name of the script being run.
        """
        super().__init__(subparser, script_name)

    def _setup_args(self, arg_parser):
        name = self._script_name
        default_dir = AMRDatabasesManager.get_default_database_directory()
        epilog = ("Example:\n"
                  "\t" + name + " info\n"
                                "\t\tPrints information about the default database in " + default_dir + "\n\n" +
                  "\t" + name + " info databases\n" +
                  "\t\tPrints information on the database stored in databases/")
        arg_parser = self._subparser.add_parser('info',
                                                epilog=epilog,
                                                formatter_class=argparse.RawTextHelpFormatter,
                                                help='Prints information on databases in the given directories.')
        arg_parser.add_argument('directories', nargs='*')

        return arg_parser

    def run(self, args):
        super(Info, self).run(args)

        arg_drug_table = ARGDrugTable()

        if len(args.directories) == 0:
            database_repos = AMRDatabasesManager.create_default_manager().get_database_repos()
            if not AMRDatabasesManager.is_database_repos_default_commits(database_repos):
                logger.warning(
                    "Using non-default ResFinder/PointFinder/PlasmidFinder database versions. This may lead to differences in the detected " +
                    "AMR genes depending on how the database files are structured.")

            try:
                database_info = database_repos.info()
                database_info.update(arg_drug_table.get_resistance_table_info())
                sys.stdout.write(get_string_with_spacing(database_info))
            except DatabaseNotFoundException as e:
                logger.error("No database found. Perhaps try restoring the default with 'staramr db restore-default'")
        else:
            for directory in args.directories:
                try:
                    database_repos = AMRDatabasesManager(directory).get_database_repos()
                    if not AMRDatabasesManager.is_database_repos_default_commits(database_repos):
                        logger.warning(
                            "Using non-default ResFinder/PointFinder/PlasmidFinder database version [%s]. This may lead to " +
                            "differences in the detected AMR genes depending on how the database files are structured.",
                            directory)

                    database_info = database_repos.info()
                    database_info.update(arg_drug_table.get_resistance_table_info())
                    sys.stdout.write(get_string_with_spacing(database_info))
                except DatabaseNotFoundException as e:
                    logger.error("Database not found in [%s]. Perhaps try building with 'staramr db build --dir %s'",
                                 directory, directory)



import logging
import tempfile
import unittest
from os import path

from Bio import SeqIO

from staramr.blast.BlastHandler import BlastHandler
from staramr.blast.plasmidfinder.PlasmidfinderBlastDatabase import PlasmidfinderBlastDatabase
from staramr.blast.resfinder.ResfinderBlastDatabase import ResfinderBlastDatabase
from staramr.databases.AMRDatabasesManager import AMRDatabasesManager
from staramr.databases.resistance.pointfinder.ARGDrugTablePointfinder import ARGDrugTablePointfinder
from staramr.databases.resistance.resfinder.ARGDrugTableResfinder import ARGDrugTableResfinder
from staramr.detection.AMRDetectionResistance import AMRDetectionResistance

logger = logging.getLogger('AMRDetectionPlasmid')


class AMRDetectionPlasmid(unittest.TestCase):

    def setUp(self):
        blast_databases_repositories = AMRDatabasesManager.create_default_manager().get_database_repos()
        self.resfinder_dir = blast_databases_repositories.get_repo_dir(
            'resfinder')
        self.pointfinder_dir = blast_databases_repositories.get_repo_dir(
            'pointfinder')
        self.plasmidfinder_dir = blast_databases_repositories.get_repo_dir(
            'plasmidfinder')

        self.resfinder_database = ResfinderBlastDatabase(self.resfinder_dir)
        self.resfinder_drug_table = ARGDrugTableResfinder()
        self.pointfinder_drug_table = ARGDrugTablePointfinder()
        self.plasmidfinder_database = PlasmidfinderBlastDatabase(
            self.plasmidfinder_dir)
        self.pointfinder_database = None
        self.blast_out = tempfile.TemporaryDirectory()
        self.blast_handler = BlastHandler(
            {'resfinder': self.resfinder_database, 'pointfinder': self.pointfinder_database,
             'plasmidfinder': self.plasmidfinder_database}, 2, self.blast_out.name)

        self.outdir = tempfile.TemporaryDirectory()
        self.amr_detection = AMRDetectionResistance(self.resfinder_database, self.resfinder_drug_table,
                                                    self.blast_handler, self.pointfinder_drug_table,
                                                    self.pointfinder_database, output_dir=self.outdir.name)

        self.test_data_dir = path.join(path.dirname(__file__), '..', 'data')

    def tearDown(self):
        self.blast_out.cleanup()
        self.outdir.cleanup()

    def testPlasmidfinderNameSuccess(self):
        file = path.join(self.test_data_dir, "test-plasmids-seq.fsa")
        files = [file]
        self.amr_detection.run_amr_detection(files, 99, 90, 90, 90)

        plasmidfinder_results = self.amr_detection.get_plasmidfinder_results()
        self.assertEqual(len(plasmidfinder_results.index), 1, 'Wrong number of rows in result')

        result = plasmidfinder_results[plasmidfinder_results['Gene'] == "IncW"]

        self.assertEqual(len(result.index), 1, 'Wrong number of results detected')
        self.assertAlmostEqual(result['%Identity'].iloc[0], 100.00, places=2, msg='Wrong pid')
        self.assertAlmostEqual(result['%Overlap'].iloc[0], 100.00, places=2, msg='Wrong overlap')
        self.assertEqual(result['Accession'].iloc[0], 'EF633507', msg='Wrong accession')
        self.assertEqual(result['HSP Length/Total Length'].iloc[0], '243/243', msg='Wrong lengths')

        hit_file = path.join(self.outdir.name, 'plasmidfinder_test-plasmids-seq.fsa')
        records = SeqIO.to_dict(SeqIO.parse(hit_file, 'fasta'))

        self.assertEqual(len(records), 1, 'Wrong number of hit records')

        expected_records = SeqIO.to_dict(SeqIO.parse(file, 'fasta'))
        self.assertEqual(expected_records['IncW_1__EF633507'].seq, records['IncW_1__EF633507'].seq,
                         "records don't match")

    def testDetailedSummary_ResPlasmid(self):
        file = path.join(self.test_data_dir, "test-detailed-summary.fsa")
        files = [file]
        self.amr_detection.run_amr_detection(files, 99, 90, 90, 90)

        detailed_summary_results = self.amr_detection.get_detailed_summary_results()
        self.assertEqual(len(detailed_summary_results.index), 2, 'Wrong number of rows in result')

        plasmid_type = detailed_summary_results[detailed_summary_results['Gene'] == "rep1"]
        self.assertEqual(len(plasmid_type.index), 1, 'Wrong number of results detected')
        self.assertEqual(plasmid_type['Predicted Phenotype'].iloc[0], '', msg='Wrong predicted phenotype')
        self.assertAlmostEqual(plasmid_type['%Identity'].iloc[0], 100.00, places=2, msg='Wrong pid')
        self.assertAlmostEqual(plasmid_type['%Overlap'].iloc[0], 100.00, places=2, msg='Wrong overlap')
        self.assertEqual(plasmid_type['Accession'].iloc[0], 'X64695', msg='Wrong accession')
        self.assertEqual(plasmid_type['HSP Length/Total Length'].iloc[0], '1491/1491', msg='Wrong lengths')
        self.assertEqual(plasmid_type['Data Type'].iloc[0], 'Plasmid', msg='Wrong data type')

        res_type = detailed_summary_results[detailed_summary_results['Gene'] == "tet(47)"]
        self.assertEqual(len(res_type.index), 1, 'Wrong number of results detected')
        self.assertEqual(res_type['Predicted Phenotype'].iloc[0], 'tetracycline', msg='Wrong predicted phenotype')
        self.assertAlmostEqual(res_type['%Identity'].iloc[0], 100.00, places=2, msg='Wrong pid')
        self.assertAlmostEqual(res_type['%Overlap'].iloc[0], 100.00, places=2, msg='Wrong overlap')
        self.assertEqual(res_type['Accession'].iloc[0], 'KR857681', msg='Wrong accession')
        self.assertEqual(res_type['HSP Length/Total Length'].iloc[0], '1248/1248', msg='Wrong lengths')
        self.assertEqual(res_type['Data Type'].iloc[0], 'Resistance', msg='Wrong data type')

    def testResistancePlasmidGenesSummary(self):
        file = path.join(self.test_data_dir, "test-resistance-plasmid.fsa")
        files = [file]
        self.amr_detection.run_amr_detection(files, 99, 90, 90, 90)

        summary_results = self.amr_detection.get_summary_results()

        self.assertEqual(len(summary_results.index), 1, 'Wrong number of rows in result')

        result = summary_results[summary_results['Genotype'] == "blaIMP-42"]
        self.assertEqual(len(result.index), 1, 'Wrong number of results detected')
        self.assertEqual(result['Predicted Phenotype'].iloc[0],
                         'ampicillin, amoxicillin/clavulanic acid, cefoxitin, ceftriaxone, meropenem',
                         msg='Wrong Predicted Phenotype')
        self.assertEqual(result['Plasmid Genes'].iloc[0], 'IncW', msg='Wrong Plasmid Gene')

    def testIndexRangePlasmids(self):
        file = path.join(self.test_data_dir, "test-index-range-plasmid.fsa")
        files = [file]
        self.amr_detection.run_amr_detection(files, 99, 90, 90, 90)

        summary_results = self.amr_detection.get_summary_results()

        self.assertEqual(len(summary_results.index), 1, 'Wrong number of rows')

        self.assertEqual(summary_results['Genotype'].iloc[0], 'None', msg='Wrong Genotype value')
        self.assertEqual(summary_results['Predicted Phenotype'].iloc[0], 'Sensitive',
                         msg='Wrong Predicted Phenotype value')
        self.assertEqual(summary_results['Plasmid Genes'].iloc[0], 'IncFII(pKPX1)', msg='Wrong Plasmid Gene')



"""
An Exception to be raised when invalid blast positions are encountered.
"""


class InvalidPositionException(Exception):

    def __init__(self, msg):
        """
        Constructs a new InvalidPositionException
        :param msg: The Exception message.
        """
        super().__init__(msg)



import copy
import logging
import os
from collections import Counter
from typing import List, Dict, Optional

from Bio import SeqIO
from pandas import DataFrame

from staramr.blast.plasmidfinder.PlasmidfinderBlastDatabase import PlasmidfinderBlastDatabase
from staramr.blast.pointfinder.PointfinderBlastDatabase import PointfinderBlastDatabase
from staramr.blast.resfinder.ResfinderBlastDatabase import ResfinderBlastDatabase
from staramr.blast.results.BlastResultsParser import BlastResultsParser
from staramr.blast.results.plasmidfinder.BlastResultsParserPlasmidfinder import BlastResultsParserPlasmidfinder
from staramr.blast.results.pointfinder.BlastResultsParserPointfinder import BlastResultsParserPointfinder
from staramr.blast.results.resfinder.BlastResultsParserResfinder import BlastResultsParserResfinder
from staramr.results.AMRDetectionSummary import AMRDetectionSummary

logger = logging.getLogger("AMRDetection")

"""
A Class to handle scanning files for AMR genes.
"""


class AMRDetection:

    def __init__(self, resfinder_database: ResfinderBlastDatabase, amr_detection_handler,
                 pointfinder_database: PointfinderBlastDatabase = None,
                 include_negative_results: bool = False, output_dir: str = None, genes_to_exclude: list = [],
                 plasmidfinder_database: PlasmidfinderBlastDatabase = None) -> None:
        """
        Builds a new AMRDetection object.
        :param resfinder_database: The staramr.blast.resfinder.ResfinderBlastDatabase for the particular ResFinder database.
        :param amr_detection_handler: The staramr.blast.BlastHandler to use for scheduling BLAST jobs.
        :param pointfinder_database: The staramr.blast.pointfinder.PointfinderBlastDatabase to use for the particular PointFinder database.
        :param plasmidfinder_database: The staramr.blast.plasmidfinder.PlasmidfinderBlastDatabase for the particular PlasmidFinder database.
        :param include_negative_results:  If True, include files lacking AMR genes in the resulting summary table.
        :param output_dir: The directory where output fasta files are to be written into (None for no output fasta files).
        :param genes_to_exclude: A list of gene IDs to exclude from the results.
        """
        self._resfinder_database = resfinder_database
        self._amr_detection_handler = amr_detection_handler
        self._pointfinder_database = pointfinder_database
        self._plasmidfinder_database = plasmidfinder_database
        self._include_negative_results = include_negative_results

        if pointfinder_database is None:
            self._has_pointfinder = False
        else:
            self._has_pointfinder = True

        self._output_dir = output_dir

        self._genes_to_exclude = genes_to_exclude

    def _create_amr_summary(self, files: List[str], resfinder_dataframe: DataFrame,
                            pointfinder_dataframe: Optional[BlastResultsParserPointfinder],
                            plasmidfinder_dataframe: DataFrame) -> DataFrame:
        amr_detection_summary = AMRDetectionSummary(files, resfinder_dataframe,
                                                    pointfinder_dataframe, plasmidfinder_dataframe)
        return amr_detection_summary.create_summary(self._include_negative_results)

    def _create_detailed_amr_summary(self, files: List[str], resfinder_dataframe: DataFrame,
                                     pointfinder_dataframe: Optional[BlastResultsParserPointfinder],
                                     plasmidfinder_dataframe: DataFrame) -> DataFrame:
        amr_detection_summary = AMRDetectionSummary(files, resfinder_dataframe,
                                                    pointfinder_dataframe, plasmidfinder_dataframe)
        return amr_detection_summary.create_detailed_summary(self._include_negative_results)

    def _create_resfinder_dataframe(self, resfinder_blast_map: Dict, pid_threshold: float, plength_threshold: int,
                                    report_all: bool) -> DataFrame:
        resfinder_parser = BlastResultsParserResfinder(resfinder_blast_map, self._resfinder_database, pid_threshold,
                                                       plength_threshold, report_all, output_dir=self._output_dir,
                                                       genes_to_exclude=self._genes_to_exclude)
        return resfinder_parser.parse_results()

    def _create_pointfinder_dataframe(self, pointfinder_blast_map: Dict, pid_threshold: float, plength_threshold: int,
                                      report_all: bool) -> DataFrame:
        pointfinder_parser = BlastResultsParserPointfinder(pointfinder_blast_map, self._pointfinder_database,
                                                           pid_threshold, plength_threshold, report_all,
                                                           output_dir=self._output_dir,
                                                           genes_to_exclude=self._genes_to_exclude)
        return pointfinder_parser.parse_results()

    def _create_plasmidfinder_dataframe(self, plasmidfinder_blast_map: Dict[str, BlastResultsParser],
                                        pid_threshold: float, plength_threshold: int,
                                        report_all: bool) -> DataFrame:
        plasmidfinder_parser = BlastResultsParserPlasmidfinder(plasmidfinder_blast_map, self._plasmidfinder_database,
                                                               pid_threshold,
                                                               plength_threshold, report_all,
                                                               output_dir=self._output_dir,
                                                               genes_to_exclude=self._genes_to_exclude)
        return plasmidfinder_parser.parse_results()

    def run_amr_detection(self, files, pid_threshold, plength_threshold_resfinder, plength_threshold_pointfinder,
                          plength_threshold_plasmidfinder, report_all=False, ignore_invalid_files=False) -> None:
        """
        Scans the passed files for AMR genes.
        :param files: The files to scan.
        :param pid_threshold: The percent identity threshold for BLAST results.
        :param plength_threshold_resfinder: The percent length overlap for BLAST results (resfinder).
        :param plength_threshold_pointfinder: The percent length overlap for BLAST results (pointfinder).
        :param plength_threshold_plasmidfinder: The percent length overlap for BLAST results (plasmidfinder).
        :param report_all: Whether or not to report all blast hits.
        :param ignore_invalid_files: Skips the invalid input files if set.
        :return: None
        """

        files_copy = copy.deepcopy(files)
        files = self._validate_files(files_copy, ignore_invalid_files)

        self._amr_detection_handler.run_blasts(files)

        resfinder_blast_map = self._amr_detection_handler.get_resfinder_outputs()
        self._resfinder_dataframe = self._create_resfinder_dataframe(resfinder_blast_map, pid_threshold,
                                                                     plength_threshold_resfinder, report_all)

        plasmidfinder_blast_map = self._amr_detection_handler.get_plasmidfinder_outputs()
        self._plasmidfinder_dataframe = self._create_plasmidfinder_dataframe(plasmidfinder_blast_map, pid_threshold,
                                                                             plength_threshold_plasmidfinder,
                                                                             report_all)

        self._pointfinder_dataframe = None
        if self._has_pointfinder:
            pointfinder_blast_map = self._amr_detection_handler.get_pointfinder_outputs()
            self._pointfinder_dataframe = self._create_pointfinder_dataframe(pointfinder_blast_map, pid_threshold,
                                                                             plength_threshold_pointfinder, report_all)

        self._summary_dataframe = self._create_amr_summary(files, self._resfinder_dataframe,
                                                           self._pointfinder_dataframe, self._plasmidfinder_dataframe)

        self._detailed_summary_dataframe = self._create_detailed_amr_summary(files, self._resfinder_dataframe,
                                                                             self._pointfinder_dataframe,
                                                                             self._plasmidfinder_dataframe)

    def _validate_files(self, files: List[str], ignore_invalid_files: bool) -> List[str]:
        total_files = len(files)
        removeable_files = []

        for file in files:
            # Check if the file is not a directory
            if os.path.isdir(file):
                if ignore_invalid_files:
                    logger.warning('--ignore-invalid-files is set, skipping directory {}'.format(file))
                    removeable_files.append(file)
                else:
                    raise Exception(
                        'Directory {} is invalid, please use --ignore-invalid-files to skip over this directory'.format(
                            file))
            else:
                # Will raise an error if the input returns an empty generator on non-FASTA files, returns a boolean
                validInput = any(SeqIO.parse(file, "fasta"))

                if not validInput:
                    if ignore_invalid_files:
                        logger.warning('--ignore-invalid-files is set, skipping file {}'.format(file))
                        removeable_files.append(file)
                    else:
                        raise Exception(
                            'File {} is invalid, please use --ignore-invalid-files to skip over invalid input files'.format(
                                file))
                else:
                    # Check if there are any duplicate sequence id's in the valid files
                    record = []
                    # Store all the sequence id's in a list
                    for sequence in SeqIO.parse(file, "fasta"):
                        record.append(sequence.id)

                    duplicates = []

                    # Each sequence contains a tuple (sequence id, frequency)
                    for sequence in (Counter(record)).items():
                        if sequence[1] > 1:
                            # We want the sequence id's that are duplicates
                            duplicates.append(sequence[0])

                    # Raise an error if there's any duplicates in the file
                    if len(duplicates) > 0:
                        raise Exception(
                            'File {} contains the following duplicate sequence IDs: {}'.format(file, duplicates))

        # Check to see if the invalid file is not the only file in the directory
        if total_files == len(removeable_files):
            raise Exception('Cannot produce output due to no valid input files')

        # Remove the skipped files
        if ignore_invalid_files:
            for file in removeable_files:
                files.remove(file)

        return files

    def get_resfinder_results(self):
        """
        Gets a pd.DataFrame for the ResFinder results.
        :return: A pd.DataFrame for the ResFinder results.
        """
        return self._resfinder_dataframe

    def get_pointfinder_results(self):
        """
        Gets a pd.DataFrame for the PointFinder results.
        :return: A pd.DataFrame for the PointFinder results.
        """
        return self._pointfinder_dataframe

    def get_plasmidfinder_results(self):
        """
        Gets a pd.DataFrame for the PlasmidFinder results.
        :return: A pd.DataFrame for the PlasmidFinder results.
        """
        return self._plasmidfinder_dataframe

    def get_summary_results(self):
        """
        Gets a pd.DataFrame for a summary table of the results.
        :return: A pd.DataFrame for a summary table of the results.
        """
        return self._summary_dataframe

    def get_detailed_summary_results(self):
        """
        Gets a pd.DataFrame for a detailed summary table of the results.
        :return: A pd.DataFrame for a detailed summary table of the results.
        """

        return self._detailed_summary_dataframe



"""
An Exception to be raised when an error is encountered with the database.
"""


class DatabaseErrorException(Exception):

    def __init__(self, msg):
        """
        Constructs a new DatabaseErrorException
        :param msg: The Exception message.
        """
        super().__init__(msg)



import logging
from os import path

from staramr.databases.resistance.ARGDrugTable import ARGDrugTable

logger = logging.getLogger("ARGDrugTableResfinder")

"""
A Class used to load up and search a file containing gene/drug mappings for ResFinder results.
"""


class ARGDrugTableResfinder(ARGDrugTable):
    DEFAULT_FILE = path.join(ARGDrugTable.DEFAULT_DATA_DIR, 'ARG_drug_key_resfinder.tsv')

    def __init__(self, file=DEFAULT_FILE):
        """
        Builds a new ARGDrugTableResfinder from the given file.
        :param file: The file containing the gene/drug mappings.
        """
        super().__init__(file=file)

    def get_drug(self, drug_class, gene_plus_variant, accession):
        """
        Gets the drug given the drug class, gene (plus variant of gene encoded in ResFinder database) and accession.
        :param drug_class: The drug class.
        :param gene_plus_variant: The gene plus variant (e.g., {gene}_{variant} = {blaIMP-58}_{1}).
        :param accession: The accession in the resfinder database (e.g., KU647281).
        :return: The particular drug, or None if no matching drug was found.
        """
        table = self._data

        drug = table[(table['Class'] == drug_class) & (table['Gene'] == gene_plus_variant) & (
                table['Accession'] == accession)]['Drug']
        if (drug.empty):
            logger.warning("No drug found for drug_class=%s, gene=%s, accession=%s", drug_class, gene_plus_variant,
                           accession)
            return None
        else:
            return self._drug_string_to_correct_separators(drug.iloc[0])



from os import path, listdir

import pandas as pd

from staramr.blast.AbstractBlastDatabase import AbstractBlastDatabase
from staramr.blast.pointfinder.PointfinderDatabaseInfo import PointfinderDatabaseInfo

"""
A Class for a PointFinder database for a particular organism.
"""


class PointfinderBlastDatabase(AbstractBlastDatabase):
    def __init__(self, database_dir, organism):
        """
        Creates a new PointfinderBlastDatabase.
        :param database_dir: The root directory for the PointFinder organisms.
        :param organism: The particular organism.
        """
        super().__init__(database_dir)
        self.organism = organism
        self.pointfinder_database_dir = path.join(self.database_dir, organism)

        if (not path.isdir(self.pointfinder_database_dir)):
            raise Exception(
                "Error, pointfinder organism [" + organism + "] is either incorrect or pointfinder database not installed properly")
        elif organism not in PointfinderBlastDatabase.get_organisms(database_dir):
            raise Exception("Pointfinder organism [" + organism + "] is not valid")

        self._pointfinder_info = PointfinderDatabaseInfo.from_file(
            path.join(self.pointfinder_database_dir, "resistens-overview.txt"))

    def get_database_names(self):
        return [f[:-len(self.fasta_suffix)] for f in listdir(self.pointfinder_database_dir) if
                (path.isfile(path.join(self.pointfinder_database_dir, f)) and f.endswith(self.fasta_suffix))]

    def get_path(self, database_name):
        return path.join(self.pointfinder_database_dir, database_name + self.fasta_suffix)

    def get_resistance_codons(self, gene, codon_mutations):
        """
        Gets a list of resistance codons from the given gene and codon mutations.
        :param gene: The gene.
        :param codon_mutations: The codon mutations.
        :return: The resistance codons.
        """
        return self._pointfinder_info.get_resistance_codons(gene, codon_mutations)

    def get_phenotype(self, gene, codon_mutation):
        """
        Gets the phenotype for a given gene and codon mutation from PointFinder.
        :param gene: The gene.
        :param codon_mutation: The codon mutation.
        :return: A string describing the phenotype.
        """
        return self._pointfinder_info.get_phenotype(gene, codon_mutation)

    def get_resistance_nucleotides(self, gene, nucleotide_mutations):
        """
        Gets a list of resistance nucleotides from the given gene and nucleotide mutations.
        :param gene: The gene.
        :param nucleotide_mutations: The nucleotide mutations.
        :return: The resistance nucleotides.
        """
        return self._pointfinder_info.get_resistance_nucleotides(gene, nucleotide_mutations)

    def get_organism(self):
        """
        Gets the particular organism of this database.
        :return: The organism.
        """
        return self.organism

    def get_name(self):
        return 'pointfinder'

    @classmethod
    def get_available_organisms(cls):
        """
        A Class Method to get a list of organisms that are currently supported by staramr.
        :return: The list of organisms currently supported by staramr.
        """
        return ['salmonella', 'campylobacter']

    @classmethod
    def get_organisms(cls, database_dir):
        """
        A Class Method to get the list of organisms from the PointFinder database root directory.
        :param database_dir: The PointFinder database root directory.
        :return: A list of organisms.
        """
        config = pd.read_csv(path.join(database_dir, 'config'), sep='\t', comment='#', header=None,
                             names=['db_prefix', 'name', 'description'])
        return config['db_prefix'].tolist()

    @classmethod
    def build_databases(cls, database_dir):
        """
        A Class Method to build a list of PointfinderBlastDatabase for all organisms in PointFinder.
        :param database_dir: The root PointFinder database directory.
        :return: A list of PointfinderBlastDatabase objects for all organisms in PointFinder.
        """
        return [PointfinderBlastDatabase(database_dir, organism) for organism in cls.get_organisms(database_dir)]



import logging
from os import path
from typing import List

import pandas as pd

from staramr.blast.AbstractBlastDatabase import AbstractBlastDatabase

logger = logging.getLogger('PlasmidfinderBlastDatabase')

"""
A Class for pulling information from the PlasmidFinder database.
"""


class PlasmidfinderBlastDatabase(AbstractBlastDatabase):

    def __init__(self, database_dir: str, database_type: str = None) -> None:
        """
        Creates a new PlasmidfinderBlastDatabase.
        :param database_dir: The specific PlasmidFinder database directory.
        """

        super().__init__(database_dir)
        self.database_type = database_type

        if database_type is not None:
            self.plasmidfinder_database_file = path.join(self.database_dir, database_type)
            self.plasmidfinder_database_file += self.fasta_suffix

            if (not path.isfile(self.plasmidfinder_database_file)):
                raise Exception(
                    "Error, plasmidfinder database type [{}] is either incorrect or plasmidfinder database not installed properly",
                    database_type)
            elif database_type not in PlasmidfinderBlastDatabase.get_database_types(database_dir):
                raise Exception("Plasmidfinder database type [{}] is not valid", database_type)

    def get_name(self) -> str:
        return 'plasmidfinder'

    def get_database_names(self) -> List[str]:
        if self.database_type is not None:
            names = [self.database_type]
        else:
            names = super().get_database_names()

        return names

    @classmethod
    def get_available_databases(self) -> list:
        """
        A Class Method to get a list of plasmidfinder databases that are currently supported by staramr.
        :return: The list of database_type currently supported by staramr.
        """
        return ['gram_positive', 'enterobacteriaceae']

    @classmethod
    def get_database_types(self, database_dir: str) -> list:
        """
        A Class Method to get the list of databases from the PlasmidFinder database root directory.
        :param database_dir: The PlasmidFinder database root directory.
        :return: A list of databases in Plasmidfinder.
        """
        config = pd.read_csv(path.join(database_dir, 'config'), sep='\t', comment='#', header=None,
                             names=['db_prefix', 'name', 'description'])

        return config['db_prefix'].tolist()



import unittest
from unittest.mock import MagicMock

from staramr.blast.results.AMRHitHSP import AMRHitHSP
from staramr.blast.results.BlastHitPartitions import BlastHitPartitions
from staramr.exceptions.InvalidPositionException import InvalidPositionException


class BlastHitPartitionsTest(unittest.TestCase):

    def testSinglePartition(self):
        hit1 = AMRHitHSP(None, None)
        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(1, len(return_list[0]), "Should only be one hit")
        self.assertEqual('contig1', return_list[0][0].get_genome_contig_id(), "Should have correct contig name")
        self.assertEqual(1, return_list[0][0].get_genome_contig_start(), "Should have correct contig start")
        self.assertEqual(10, return_list[0][0].get_genome_contig_end(), "Should have correct contig end")

    def testSinglePartitionMinus(self):
        hit1 = AMRHitHSP(None, None)
        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=10)
        hit1.get_genome_contig_end = MagicMock(return_value=1)
        hit1.get_genome_contig_strand = MagicMock(return_value='minus')

        parts = BlastHitPartitions()

        parts.append(hit1)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(1, len(return_list[0]), "Should only be one hit")
        self.assertEqual('contig1', return_list[0][0].get_genome_contig_id(), "Should have correct contig name")
        self.assertEqual(10, return_list[0][0].get_genome_contig_start(), "Should have correct contig start")
        self.assertEqual(1, return_list[0][0].get_genome_contig_end(), "Should have correct contig end")

    def testSinglePartitionPlusFailMinusCoords(self):
        hit1 = AMRHitHSP(None, None)
        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=10)
        hit1.get_genome_contig_end = MagicMock(return_value=1)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        self.assertRaises(InvalidPositionException, parts.append, hit1)

    def testSinglePartitionIdenticalHits(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit1)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([1, 1], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 10], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionIdenticalHitsMinusStrand(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=10)
        hit1.get_genome_contig_end = MagicMock(return_value=1)
        hit1.get_genome_contig_strand = MagicMock(return_value='minus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit1)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([10, 10], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([1, 1], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionSameLocationOppositeStrands(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=10)
        hit1.get_genome_contig_end = MagicMock(return_value=1)
        hit1.get_genome_contig_strand = MagicMock(return_value='minus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=1)
        hit2.get_genome_contig_end = MagicMock(return_value=10)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([10, 1], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([1, 10], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHitEndGreater(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=1)
        hit2.get_genome_contig_end = MagicMock(return_value=11)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([1, 1], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 11], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHitEndGreaterMinus(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=11)
        hit2.get_genome_contig_end = MagicMock(return_value=1)
        hit2.get_genome_contig_strand = MagicMock(return_value='minus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([1, 11], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 1], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHitEndGreaterStartGreater(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=2)
        hit2.get_genome_contig_end = MagicMock(return_value=11)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([1, 2], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 11], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHitEndGreaterStartGreaterMinus(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=11)
        hit2.get_genome_contig_end = MagicMock(return_value=2)
        hit2.get_genome_contig_strand = MagicMock(return_value='minus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([1, 11], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 2], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHitEndLesser(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=1)
        hit2.get_genome_contig_end = MagicMock(return_value=9)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([1, 1], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 9], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHitEndLesserMinus(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=9)
        hit2.get_genome_contig_end = MagicMock(return_value=1)
        hit2.get_genome_contig_strand = MagicMock(return_value='minus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([1, 9], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 1], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHitEndLesserStartLesser(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=2)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=1)
        hit2.get_genome_contig_end = MagicMock(return_value=9)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([2, 1], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 9], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHitEndLesserStartLesserMinus(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=2)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=9)
        hit2.get_genome_contig_end = MagicMock(return_value=1)
        hit2.get_genome_contig_strand = MagicMock(return_value='minus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([2, 9], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 1], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHitEndGreaterStartLesser(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=2)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=1)
        hit2.get_genome_contig_end = MagicMock(return_value=11)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([2, 1], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 11], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHitEndGreaterStartLesserMinus(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=2)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=11)
        hit2.get_genome_contig_end = MagicMock(return_value=1)
        hit2.get_genome_contig_strand = MagicMock(return_value='minus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([2, 11], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 1], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHit2InHit1(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=2)
        hit2.get_genome_contig_end = MagicMock(return_value=9)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([1, 2], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 9], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHit2InHit1Minus(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=9)
        hit2.get_genome_contig_end = MagicMock(return_value=2)
        hit2.get_genome_contig_strand = MagicMock(return_value='minus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([1, 9], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 2], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHit2EdgeWithinHit1Lesser(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=5)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=1)
        hit2.get_genome_contig_end = MagicMock(return_value=6)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([5, 1], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10, 6], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHit2EdgeSameHit1Lesser(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=5)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=1)
        hit2.get_genome_contig_end = MagicMock(return_value=5)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(2, len(return_list), "Should be two partitions")
        self.assertEqual(1, len(return_list[0]), "Partition 1 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([5], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10], [x.get_genome_contig_end() for x in return_list[0]], "Should have correct contig ends")

        self.assertEqual(1, len(return_list[1]), "Partition 2 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[1]],
                         "Should have correct contig names")
        self.assertEqual([1], [x.get_genome_contig_start() for x in return_list[1]],
                         "Should have correct contig starts")
        self.assertEqual([5], [x.get_genome_contig_end() for x in return_list[1]], "Should have correct contig ends")

    def testSinglePartitionHit2EdgeWithinHit1Greater(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=5)
        hit1.get_genome_contig_end = MagicMock(return_value=11)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=10)
        hit2.get_genome_contig_end = MagicMock(return_value=15)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(1, len(return_list), "Should only be one partition")
        self.assertEqual(2, len(return_list[0]), "Should be two hits")
        self.assertEqual(['contig1', 'contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([5, 10], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([11, 15], [x.get_genome_contig_end() for x in return_list[0]],
                         "Should have correct contig ends")

    def testSinglePartitionHit2EdgeSameHit1Greater(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=5)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=10)
        hit2.get_genome_contig_end = MagicMock(return_value=15)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(2, len(return_list), "Should be two partitions")
        self.assertEqual(1, len(return_list[0]), "Partition 1 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([5], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10], [x.get_genome_contig_end() for x in return_list[0]], "Should have correct contig ends")

        self.assertEqual(1, len(return_list[1]), "Partition 2 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[1]],
                         "Should have correct contig names")
        self.assertEqual([10], [x.get_genome_contig_start() for x in return_list[1]],
                         "Should have correct contig starts")
        self.assertEqual([15], [x.get_genome_contig_end() for x in return_list[1]], "Should have correct contig ends")

    def testTwoPartitionsHit2EdgeHit1Lesser(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=5)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=1)
        hit2.get_genome_contig_end = MagicMock(return_value=4)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(2, len(return_list), "Should be two partitions")
        self.assertEqual(1, len(return_list[0]), "Partition 1 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([5], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10], [x.get_genome_contig_end() for x in return_list[0]], "Should have correct contig ends")

        self.assertEqual(1, len(return_list[1]), "Partition 2 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[1]],
                         "Should have correct contig names")
        self.assertEqual([1], [x.get_genome_contig_start() for x in return_list[1]],
                         "Should have correct contig starts")
        self.assertEqual([4], [x.get_genome_contig_end() for x in return_list[1]], "Should have correct contig ends")

    def testTwoPartitionsHit2EdgeHit1Greater(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=5)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=11)
        hit2.get_genome_contig_end = MagicMock(return_value=15)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(2, len(return_list), "Should be two partitions")
        self.assertEqual(1, len(return_list[0]), "Partition 1 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([5], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10], [x.get_genome_contig_end() for x in return_list[0]], "Should have correct contig ends")

        self.assertEqual(1, len(return_list[1]), "Partition 2 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[1]],
                         "Should have correct contig names")
        self.assertEqual([11], [x.get_genome_contig_start() for x in return_list[1]],
                         "Should have correct contig starts")
        self.assertEqual([15], [x.get_genome_contig_end() for x in return_list[1]], "Should have correct contig ends")

    def testTwoPartitionsHit2EdgeHit1GreaterMinus(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=5)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig1")
        hit2.get_genome_contig_start = MagicMock(return_value=15)
        hit2.get_genome_contig_end = MagicMock(return_value=11)
        hit2.get_genome_contig_strand = MagicMock(return_value='minus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(2, len(return_list), "Should be two partitions")
        self.assertEqual(1, len(return_list[0]), "Partition 1 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([5], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10], [x.get_genome_contig_end() for x in return_list[0]], "Should have correct contig ends")

        self.assertEqual(1, len(return_list[1]), "Partition 2 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[1]],
                         "Should have correct contig names")
        self.assertEqual([15], [x.get_genome_contig_start() for x in return_list[1]],
                         "Should have correct contig starts")
        self.assertEqual([11], [x.get_genome_contig_end() for x in return_list[1]], "Should have correct contig ends")

    def testTwoPartitionsDifferentContigNames(self):
        hit1 = AMRHitHSP(None, None)

        hit1.get_genome_contig_id = MagicMock(return_value="contig1")
        hit1.get_genome_contig_start = MagicMock(return_value=1)
        hit1.get_genome_contig_end = MagicMock(return_value=10)
        hit1.get_genome_contig_strand = MagicMock(return_value='plus')

        hit2 = AMRHitHSP(None, None)

        hit2.get_genome_contig_id = MagicMock(return_value="contig2")
        hit2.get_genome_contig_start = MagicMock(return_value=1)
        hit2.get_genome_contig_end = MagicMock(return_value=10)
        hit2.get_genome_contig_strand = MagicMock(return_value='plus')

        parts = BlastHitPartitions()

        parts.append(hit1)
        parts.append(hit2)

        return_list = parts.get_hits_nonoverlapping_regions()
        self.assertEqual(2, len(return_list), "Should be two partitions")
        self.assertEqual(1, len(return_list[0]), "Partition 1 should have 1 hit")
        self.assertEqual(['contig1'], [x.get_genome_contig_id() for x in return_list[0]],
                         "Should have correct contig names")
        self.assertEqual([1], [x.get_genome_contig_start() for x in return_list[0]],
                         "Should have correct contig starts")
        self.assertEqual([10], [x.get_genome_contig_end() for x in return_list[0]], "Should have correct contig ends")

        self.assertEqual(1, len(return_list[1]), "Partition 2 should have 1 hit")
        self.assertEqual(['contig2'], [x.get_genome_contig_id() for x in return_list[1]],
                         "Should have correct contig names")
        self.assertEqual([1], [x.get_genome_contig_start() for x in return_list[1]],
                         "Should have correct contig starts")
        self.assertEqual([10], [x.get_genome_contig_end() for x in return_list[1]], "Should have correct contig ends")



import logging
from os import path

from staramr.databases.resistance.ARGDrugTable import ARGDrugTable

logger = logging.getLogger("ARGDrugTablePointfinder")

"""
A Class used to load up and search a file containing gene/drug mappings for PointFinder results.
"""


class ARGDrugTablePointfinder(ARGDrugTable):
    DEFAULT_FILE = path.join(ARGDrugTable.DEFAULT_DATA_DIR, 'ARG_drug_key_pointfinder.tsv')

    def __init__(self, file=DEFAULT_FILE):
        """
        Builds a new ARGDrugTablePointfinder from the given file.
        :param file: The file containing the gene/drug mappings.
        """
        super().__init__(file=file)

    def get_drug(self, organism, gene, position):
        """
        Gets the drug given the organism, gene, and position of point mutation.
        :param organism: The organism.
        :param gene: The gene.
        :param position: The position of the point mutation (may be codon position or nucleotide position depending on gene).
        :return: The drug this mutation causes resistance to, or None if no such drug.
        """
        table = self._data

        drug = table[(table['Organism'] == organism) & (table['Gene'] == gene) & (
                table['Codon Pos.'] == position)]['Drug']
        if (drug.empty):
            logger.warning("No drug found for organism=%s, gene=%s, position=%s", organism, gene, position)
            return None
        else:
            return self._drug_string_to_correct_separators(drug.iloc[0])



import unittest

import pandas as pd

from staramr.results.AMRDetectionSummaryResistance import AMRDetectionSummaryResistance


class AMRDetectionSummaryResistanceTest(unittest.TestCase):

    def setUp(self):
        self.columns_resfinder = ('Isolate ID', 'Gene', 'Predicted Phenotype', '%Identity', '%Overlap',
                                  'HSP Length/Total Length', 'Contig', 'Start', 'End', 'Accession')
        self.columns_pointfinder = ('Isolate ID', 'Gene', 'Predicted Phenotype', 'Type', 'Position', 'Mutation',
                                    '%Identity', '%Overlap', 'HSP Length/Total Length')

        # Resfinder tables
        self.resfinder_table_empty = pd.DataFrame([],
                                                  columns=self.columns_resfinder)

        self.resfinder_table = pd.DataFrame([
            ['file1', 'blaIMP-42', 'ampicillin, amoxi/clav, cefoxitin, ceftriaxone, meropenem', 99.73, 100.00,
             '741/741', 'blaIMP-42_1_AB753456', 1, 741, 'AB753456']
        ],
            columns=self.columns_resfinder)

        self.resfinder_table_duplicate_resistances = pd.DataFrame([
            ['file1', 'blaIMP-42', 'ampicillin', 99.73, 100.00,
             '741/741', 'blaIMP-42_1_AB753456', 1, 741, 'AB753456'],
            ['file1', 'blaCTX-M-55', 'ampicillin, ceftriaxone', 99.73, 100.00,
             '741/741', 'x', 1, 741, 'AB753456']
        ],
            columns=self.columns_resfinder)

        self.pointfinder_table = pd.DataFrame([
            ['file1', 'gyrA', 'ciprofloxacin I/R, nalidixic acid', 'codon', 67, 'GCC -> CCC (A -> P)', 99.96, 100.0,
             '2637/2637'],
        ],
            columns=self.columns_pointfinder)

        self.pointfinder_table_duplicate = pd.DataFrame([
            ['file1', 'gyrA', 'ampicillin, ceftriaxone, ciprofloxacin I/R', 'codon', 67, 'GCC -> CCC (A -> P)', 99.96,
             100.0,
             '2637/2637'],
        ],
            columns=self.columns_pointfinder)

        self.files = ['file1']

    def testResfinder(self):
        amr_detection_summary = AMRDetectionSummaryResistance(self.files, self.resfinder_table)

        summary = amr_detection_summary.create_summary()

        self.assertEqual(1, len(summary.index), 'Invalid number of rows in results')

        self.assertEqual('file1', summary.index[0], 'File name not equal')
        self.assertEqual('blaIMP-42', summary['Genotype'].iloc[0], 'Genes not equal')
        self.assertEqual('ampicillin, amoxi/clav, cefoxitin, ceftriaxone, meropenem',
                         summary['Predicted Phenotype'].iloc[0], 'Phenotypes not equal')

    def testResfinderDuplicateResistances(self):
        amr_detection_summary = AMRDetectionSummaryResistance(self.files, self.resfinder_table_duplicate_resistances)

        summary = amr_detection_summary.create_summary()

        self.assertEqual(1, len(summary.index), 'Invalid number of rows in results')

        self.assertEqual('file1', summary.index[0], 'File name not equal')
        self.assertEqual('blaCTX-M-55, blaIMP-42', summary['Genotype'].iloc[0], 'Genes not equal')
        self.assertEqual('ampicillin, ceftriaxone',
                         summary['Predicted Phenotype'].iloc[0], 'Phenotypes not equal')

    def testPointfinder(self):
        amr_detection_summary = AMRDetectionSummaryResistance(self.files, self.resfinder_table_empty,
                                                              self.pointfinder_table)

        summary = amr_detection_summary.create_summary()

        self.assertEqual(1, len(summary.index), 'Invalid number of rows in results')

        self.assertEqual('file1', summary.index[0], 'File name not equal')
        self.assertEqual('gyrA', summary['Genotype'].iloc[0], 'Genes not equal')
        self.assertEqual('ciprofloxacin I/R, nalidixic acid',
                         summary['Predicted Phenotype'].iloc[0], 'Phenotypes not equal')

    def testPointfinderResfinder(self):
        amr_detection_summary = AMRDetectionSummaryResistance(self.files, self.resfinder_table, self.pointfinder_table)

        summary = amr_detection_summary.create_summary()

        self.assertEqual(1, len(summary.index), 'Invalid number of rows in results')

        self.assertEqual('file1', summary.index[0], 'File name not equal')
        self.assertEqual('blaIMP-42, gyrA', summary['Genotype'].iloc[0], 'Genes not equal')
        self.assertEqual('ampicillin, amoxi/clav, cefoxitin, ceftriaxone, meropenem, ciprofloxacin I/R, nalidixic acid',
                         summary['Predicted Phenotype'].iloc[0], 'Phenotypes not equal')

    def testPointfinderResfinderDuplicate(self):
        amr_detection_summary = AMRDetectionSummaryResistance(self.files, self.resfinder_table_duplicate_resistances,
                                                              self.pointfinder_table_duplicate)

        summary = amr_detection_summary.create_summary()

        self.assertEqual(1, len(summary.index), 'Invalid number of rows in results')

        self.assertEqual('file1', summary.index[0], 'File name not equal')
        self.assertEqual('blaCTX-M-55, blaIMP-42, gyrA', summary['Genotype'].iloc[0], 'Genes not equal')
        self.assertEqual('ampicillin, ceftriaxone, ciprofloxacin I/R',
                         summary['Predicted Phenotype'].iloc[0], 'Phenotypes not equal')



import logging

from staramr.blast.results.pointfinder.BlastResultsParserPointfinder import BlastResultsParserPointfinder

"""
Class used to parse out BLAST results for PointFinder, including phenotyhpes/resistances.
"""

logger = logging.getLogger('BlastResultsParserPointfinderResistance')


class BlastResultsParserPointfinderResistance(BlastResultsParserPointfinder):
    COLUMNS = [x.strip() for x in '''
    Isolate ID
    Gene
    Predicted Phenotype
    Type
    Position
    Mutation
    %Identity
    %Overlap
    HSP Length/Total Length
    Contig
    Start
    End
    '''.strip().split('\n')]

    def __init__(self, file_blast_map, arg_drug_table, blast_database, pid_threshold, plength_threshold,
                 report_all=False, output_dir=None, genes_to_exclude=[]):
        """
        Creates a new BlastResultsParserPointfinderResistance.
        :param file_blast_map: A map/dictionary linking input files to BLAST results files.
        :param arg_drug_table: A table mapping the resistance gene to a specific drug resistance.
        :param blast_database: The particular staramr.blast.AbstractBlastDatabase to use.
        :param pid_threshold: A percent identity threshold for BLAST results.
        :param plength_threshold: A percent length threshold for results.
        :param report_all: Whether or not to report all blast hits.
        :param output_dir: The directory where output files are being written.
        :param genes_to_exclude: A list of gene IDs to exclude from the results.
        """
        super().__init__(file_blast_map, blast_database, pid_threshold, plength_threshold, report_all,
                         output_dir=output_dir, genes_to_exclude=genes_to_exclude)
        self._arg_drug_table = arg_drug_table

    def _get_result(self, hit, db_mutation):
        drug = self._arg_drug_table.get_drug(self._blast_database.get_organism(), hit.get_amr_gene_id(),
                                             db_mutation.get_mutation_position())
        gene_name = hit.get_amr_gene_id() + " (" + db_mutation.get_mutation_string_short() + ")"

        if drug is None:
            drug = 'unknown[' + gene_name + ']'

        return [hit.get_genome_id(),
                gene_name,
                drug,
                db_mutation.get_type(),
                db_mutation.get_mutation_position(),
                db_mutation.get_mutation_string(),
                hit.get_pid(),
                hit.get_plength(),
                str(hit.get_hsp_length()) + "/" + str(hit.get_amr_gene_length()),
                hit.get_genome_contig_id(),
                hit.get_genome_contig_start(),
                hit.get_genome_contig_end()
                ]



import unittest

from staramr.blast.results.plasmidfinder.PlasmidfinderHitHSP import PlasmidfinderHitHSP


class PlasmidfinderHitHSPTest(unittest.TestCase):

    def testParseSequenceId1(self):
        test_blast_record = {"sstart": 20, "send": 30, "sstrand": "ABC", "qstart": 1, "qend": 10,
                             'qseqid': 'RepA_1_pKPC-CAV1321_CP011611'}

        plasmid_hit_hsp = PlasmidfinderHitHSP('test_file', test_blast_record)

        self.assertEqual('RepA', plasmid_hit_hsp.get_amr_gene_name(), 'Did not parse correct gene name')
        self.assertEqual('1', plasmid_hit_hsp.get_amr_gene_variant(), 'Did not parse correct gene variant')
        self.assertEqual('RepA_1', plasmid_hit_hsp.get_amr_gene_name_with_variant(),
                         'Did not parse correct gene name variant')
        self.assertEqual('CP011611', plasmid_hit_hsp.get_amr_gene_accession(),
                         'Did not parse correct gene name variant')
        self.assertEqual('RepA_1_CP011611', plasmid_hit_hsp.get_amr_gene_variant_accession(),
                         'Did not parse correct gene name variant accession')

    def testParseSequenceId2(self):
        test_blast_record = {"sstart": 20, "send": 30, "sstrand": "ABC", "qstart": 1, "qend": 10,
                             'qseqid': 'IncHI2_1__BX664015'}

        plasmid_hit_hsp = PlasmidfinderHitHSP('test_file', test_blast_record)

        self.assertEqual('IncHI2', plasmid_hit_hsp.get_amr_gene_name(), 'Did not parse correct gene name')
        self.assertEqual('1', plasmid_hit_hsp.get_amr_gene_variant(), 'Did not parse correct gene variant')
        self.assertEqual('IncHI2_1', plasmid_hit_hsp.get_amr_gene_name_with_variant(),
                         'Did not parse correct gene name variant')
        self.assertEqual('BX664015', plasmid_hit_hsp.get_amr_gene_accession(),
                         'Did not parse correct gene name variant')
        self.assertEqual('IncHI2_1_BX664015', plasmid_hit_hsp.get_amr_gene_variant_accession(),
                         'Did not parse correct gene name variant accession')

    def testParseSequenceId3(self):
        test_blast_record = {"sstart": 20, "send": 30, "sstrand": "ABC", "qstart": 1, "qend": 10,
                             'qseqid': 'IncB/O/K/Z_1__CU928147'}

        plasmid_hit_hsp = PlasmidfinderHitHSP('test_file', test_blast_record)

        self.assertEqual('IncB/O/K/Z', plasmid_hit_hsp.get_amr_gene_name(), 'Did not parse correct gene name')
        self.assertEqual('1', plasmid_hit_hsp.get_amr_gene_variant(), 'Did not parse correct gene variant')
        self.assertEqual('IncB/O/K/Z_1', plasmid_hit_hsp.get_amr_gene_name_with_variant(),
                         'Did not parse correct gene name variant')
        self.assertEqual('CU928147', plasmid_hit_hsp.get_amr_gene_accession(),
                         'Did not parse correct gene name variant')
        self.assertEqual('IncB/O/K/Z_1_CU928147', plasmid_hit_hsp.get_amr_gene_variant_accession(),
                         'Did not parse correct gene name variant accession')

    def testParseSequenceId4(self):
        test_blast_record = {"sstart": 20, "send": 30, "sstrand": "ABC", "qstart": 1, "qend": 10,
                             'qseqid': 'IncFII(Serratia)_1_Serratia_NC_009829'}

        plasmid_hit_hsp = PlasmidfinderHitHSP('test_file', test_blast_record)

        self.assertEqual('IncFII(Serratia)', plasmid_hit_hsp.get_amr_gene_name(), 'Did not parse correct gene name')
        self.assertEqual('1', plasmid_hit_hsp.get_amr_gene_variant(), 'Did not parse correct gene variant')
        self.assertEqual('IncFII(Serratia)_1', plasmid_hit_hsp.get_amr_gene_name_with_variant(),
                         'Did not parse correct gene name variant')
        self.assertEqual('NC_009829', plasmid_hit_hsp.get_amr_gene_accession(),
                         'Did not parse correct gene name variant')
        self.assertEqual('IncFII(Serratia)_1_NC_009829', plasmid_hit_hsp.get_amr_gene_variant_accession(),
                         'Did not parse correct gene name variant accession')



from staramr.databases.resistance.pointfinder.ARGDrugTablePointfinder import ARGDrugTablePointfinder
from staramr.databases.resistance.resfinder.ARGDrugTableResfinder import ARGDrugTableResfinder
from staramr.detection.AMRDetection import AMRDetection
from staramr.detection.AMRDetectionResistance import AMRDetectionResistance

"""
A Class used to construct a particular staramr.detection.AMRDetection object.
"""


class AMRDetectionFactory:

    def __init__(self):
        pass

    def build(self, plasmidfinder_database, resfinder_database, blast_handler, pointfinder_database, include_negatives,
              include_resistances=False, output_dir=None, genes_to_exclude=[]):
        """
        Builds a new AMRDetection object.
        :param plasmidfinder_database: The staramr.blast.plasmidfinder.PlasmidfinderBlastDatabase to use for the particular PlasmidFinder database.
        :param resfinder_database: The staramr.blast.resfinder.ResfinderBlastDatabase for the particular ResFinder database.
        :param blast_handler: The staramr.blast.BlastHandler to use for scheduling BLAST jobs.
        :param pointfinder_database: The staramr.blast.pointfinder.PointfinderBlastDatabase to use for the particular PointFinder database.
        :param include_negatives:  If True, include files lacking AMR genes in the resulting summary table.
        :param include_resistances: If True, include predicted drug resistances in output.
        :param output_dir: The directory where output files are being written.
        :param genes_to_exclude: A list of gene IDs to exclude from the results.
        :return: A new AMRDetection object.
        """

        if include_resistances:
            return AMRDetectionResistance(resfinder_database, ARGDrugTableResfinder(), blast_handler,
                                          ARGDrugTablePointfinder(), pointfinder_database, include_negatives,
                                          output_dir=output_dir, genes_to_exclude=genes_to_exclude,
                                          plasmidfinder_database=plasmidfinder_database)
        else:
            return AMRDetection(resfinder_database, blast_handler, pointfinder_database, include_negatives,
                                output_dir=output_dir, genes_to_exclude=genes_to_exclude,
                                plasmidfinder_database=plasmidfinder_database)



import configparser
import logging
import shutil
import time
from collections import OrderedDict
from os import path
from typing import Dict

import git

from staramr.exceptions.DatabaseErrorException import DatabaseErrorException
from staramr.exceptions.DatabaseNotFoundException import DatabaseNotFoundException

"""
A Class used to handle interactions with the BLAST database repositories.
"""


class BlastDatabaseRepository:
    TIME_FORMAT = "%a, %d %b %Y %H:%M"
    LOGGER = logging.getLogger('BlastDatabaseRepository')

    def __init__(self, database_root_dir: str, database_name: str, git_repository_url: str) -> None:
        """
        Creates a new BlastDatabaseRepository.
        :param database_root_dir: The root directory for both the Blast databases.
        :param database_name: A name for this database.
        :param git_repository_url: A URL to the git repository managing the database files.
        """
        self._database_dir = database_root_dir
        self._database_name = database_name
        self._git_repository_url = git_repository_url

        self._git_dir = path.join(database_root_dir, database_name)

    def build(self, commit: str = None) -> None:
        """
        Downloads and builds a new Blast database.
        :param commit: The specific git commit to download. Defaults to latest commit.
        :return: None
        """

        try:
            self.LOGGER.info("Cloning %s db [%s] to [%s]", self._database_name, self._git_repository_url, self._git_dir)
            repo = git.repo.base.Repo.clone_from(self._git_repository_url, self._git_dir)

            if commit is not None:
                self.LOGGER.info("Checking out %s commit %s", self._database_name, commit)
                repo.git.checkout(commit)
        except Exception as e:
            raise DatabaseErrorException("Could not build database in [" + self._database_dir + "]") from e

    def update(self, commit: str = None) -> None:
        """
        Updates an existing Blast database to the latest revisions (or passed specific revisions).
        :param commit: The specific git commit to update to. Defaults to latest commit.
        :return: None
        """

        if not path.exists(self._git_dir):
            self.build(commit=commit)
        else:
            try:
                repo = git.Repo(self._git_dir)

                self.LOGGER.info("Updating %s", self._git_dir)
                repo.heads.master.checkout()
                repo.remotes.origin.pull()

                if commit is not None:
                    self.LOGGER.info("Checking out %s commit %s", self._database_name, commit)
                    repo.git.checkout(commit)

                    repo.git.reset('--hard')
            except Exception as e:
                raise DatabaseErrorException("Could not build database in [" + self._database_dir + "]") from e

    def remove(self):
        """
        Removes the databases stored in this directory.
        :return: None
        """
        shutil.rmtree(self._git_dir)

    def is_at_commit(self, commit: str = None) -> bool:
        """
        Determines whether this database repo is at the specified commit.
        :param commit: The commit to check.
        :return: True if the database is at the specified commit, otherwise False.
        """
        return self.info()[self._get_info_name('commit')] == commit

    def info(self) -> Dict[str, str]:
        """
        Gets information on the Blast databases.
        :return: Database information as a OrderedDict of key/value pairs.
        """
        data = OrderedDict()  # type: Dict[str,str]

        try:
            repo = git.Repo(self._git_dir)
            repo_head = repo.commit('HEAD')

            data[self._get_info_name('dir')] = self._git_dir
            data[self._get_info_name('url')] = self._git_repository_url
            data[self._get_info_name('commit')] = str(repo_head)
            data[self._get_info_name('date')] = time.strftime(self.TIME_FORMAT,
                                                              time.gmtime(repo_head.committed_date))

        except git.exc.NoSuchPathError as e:
            raise DatabaseNotFoundException('Invalid database in [' + self._database_dir + ']') from e

        return data

    def _get_info_name(self, info_type):
        return self._database_name + '_db_' + info_type

    def get_database_dir(self) -> str:
        """
        Gets the root database dir.
        :return: The root database dir.
        """
        return self._database_dir

    def get_git_dir(self) -> str:
        """
        Gets the database git directory.
        :return: The database git directory.
        """
        return self._git_dir


"""
A Class used to handle interactions with the BLAST database repositories, stripping out the .git directory.
"""


class BlastDatabaseRepositoryStripGitDir(BlastDatabaseRepository):
    GIT_INFO_SECTION = 'GitInfo'
    LOGGER = logging.getLogger('BlastDatabaseRepositoryStripGitDir')

    def __init__(self, database_root_dir: str, database_name: str, git_repository_url: str) -> None:
        """
        Creates a new BlastDatabaseRepositoryStripGitDir.
        :param database_root_dir: The root directory for both the Blast databases.
        :param database_name: A name for this database.
        :param git_repository_url: A URL to the git repository managing the database files.
        """
        super().__init__(database_root_dir, database_name, git_repository_url)

        self._git_dot_git_dir = path.join(self._git_dir, '.git')
        self._info_file = path.join(database_root_dir, database_name + '-info.ini')

    def build(self, commit: str = None):
        """
        Downloads and builds a new Blast database.
        :param commit: The specific git commit to download. Defaults to latest commit.
        :return: None
        """
        super().build(commit=commit)

        database_info = super().info()

        # remove directories from info as they are unimportant here
        database_info_stripped = OrderedDict(database_info)
        del database_info_stripped[self._get_info_name('dir')]

        self._write_database_info_to_file(database_info_stripped, self._info_file)

        self.LOGGER.info("Removing %s", self._git_dot_git_dir)
        shutil.rmtree(self._git_dot_git_dir)

    def _write_database_info_to_file(self, database_info, file):
        config = configparser.ConfigParser()
        config[self.GIT_INFO_SECTION] = database_info

        with open(file, 'w') as file_handle:
            config.write(file_handle)

    def _read_database_info_from_file(self, file):
        config = configparser.ConfigParser()
        config.read(file)
        return OrderedDict(config[self.GIT_INFO_SECTION])

    def update(self, commit: str = None) -> None:
        """
        Updates an existing Blast database to the latest revisions (or passed specific revisions).
        :param commit: The commit to update to.
        :return: None
        """
        raise Exception("Cannot update when .git directory has been removed")

    def info(self) -> Dict[str, str]:
        """
        Gets information on the ResFinder/PointFinder databases.
        :return: Database information as a list containing key/value pairs.
        """

        try:
            data = self._read_database_info_from_file(self._info_file)
            data[self._get_info_name('dir')] = self._git_dir

            # re-order all fields
            data.move_to_end(self._get_info_name('dir'), last=True)
            data.move_to_end(self._get_info_name('url'), last=True)
            data.move_to_end(self._get_info_name('commit'), last=True)
            data.move_to_end(self._get_info_name('date'), last=True)

            return data
        except FileNotFoundError as e:
            raise DatabaseNotFoundException('Database could not be found in [' + self.get_database_dir() + ']') from e



import logging
import shutil
from collections import OrderedDict
from typing import Dict

from staramr.blast.AbstractBlastDatabase import AbstractBlastDatabase
from staramr.blast.plasmidfinder.PlasmidfinderBlastDatabase import PlasmidfinderBlastDatabase
from staramr.blast.pointfinder.PointfinderBlastDatabase import PointfinderBlastDatabase
from staramr.blast.resfinder.ResfinderBlastDatabase import ResfinderBlastDatabase
from staramr.databases.BlastDatabaseRepository import BlastDatabaseRepository, BlastDatabaseRepositoryStripGitDir

logger = logging.getLogger('BlastDatabaseRepositories')

"""
A Class used to handle interactions with blast database repository files.
"""


class BlastDatabaseRepositories:

    def __init__(self, database_dir: str, is_dist: bool = False) -> None:
        """
        Creates a new AMRDatabaseHandler.
        :param database_dir: The root directory for the databases.
        :param is_dist: Whether or not we are building distributable versions of the blast database repositories
            (that is, should we strip out the .git directories).
        """
        self._database_dir = database_dir
        self._database_repositories = {}  # type: Dict[str,BlastDatabaseRepository]
        self._is_dist = is_dist

    def register_database_repository(self, database_name: str, git_repository_url: str) -> None:
        """
        Registers a new database repository.
        :param database_name: The name of the database.
        :param git_repository_url: The git repository url.
        :param is_dist: True if this database should be interpreted as the distributable version (no .git directory).
        :return: None
        """
        database_repository = BlastDatabaseRepository(self._database_dir, database_name,
                                                      git_repository_url)  # type: BlastDatabaseRepository
        if self._is_dist:
            database_repository = BlastDatabaseRepositoryStripGitDir(self._database_dir, database_name,
                                                                     git_repository_url)

        if database_name in self._database_repositories:
            raise Exception("A database with name [{}] already exists", database_name)
        else:
            self._database_repositories[database_name] = database_repository

    def build(self, commits: Dict[str, str] = None):
        """
        Downloads and builds new databases.
        :param commits: A map of {'database_name' : 'commit'} defining the particular commits to build.
        :return: None
        """
        for database_name in self._database_repositories:
            commit = commits.get(database_name) if commits else None
            self._database_repositories[database_name].build(commit)

    def update(self, commits: Dict[str, str] = None):
        """
        Updates an existing database to the latest revisions (or passed specific revisions).
        :param commits: A map of {'database_name' : 'commit'} defining the particular commits to update to.
        :return: None
        """
        for database_name in self._database_repositories:
            commit = commits.get(database_name) if commits else None
            self._database_repositories[database_name].update(commit)

    def remove(self):
        """
        Removes the databases stored in this directory.
        :return: None
        """
        for name, repo in self._database_repositories.items():
            repo.remove()

        shutil.rmtree(self._database_dir)

    def info(self) -> Dict[str, str]:
        """
        Gets information on the ResFinder/PointFinder databases.
        :return: Database information as a OrderedDict of key/value pairs.
        """
        info = OrderedDict()  # type: Dict[str,str]

        for name, repo in self._database_repositories.items():
            info.update(repo.info())

        return info

    def get_database_dir(self) -> str:
        """
        Gets the root database dir.
        :return: The root database dir.
        """
        return self._database_dir

    def get_repo_dir(self, name: str) -> str:
        """
        Gets database repo directory for the given database name.
        :param name: The database name.
        :return: The database dir for the given database name.
        """
        return self._database_repositories[name].get_git_dir()

    def is_at_commits(self, commits: Dict[str, str]):
        """
        Are the database repositories at the passed commits?
        :param commits: A dict of the commits {'database_name': 'commit'}.
        :return: True if the database repositories are at the passed commits (ignores repos not passed in dict). False otherwise.
        """
        for name, repo in self._database_repositories.items():
            if name in commits and not repo.is_at_commit(commits[name]):
                return False

        return True

    def is_dist(self):
        """
        Whether or not we are building distributable versions of the blast database repositories (that is, should we strip out the .git directories).
        :return: True if is_dist, False otherwise.
        """
        return self._is_dist

    @classmethod
    def create_default_repositories(cls, root_database_dir: str, is_dist: bool = False):
        """
        Class method for creating a BlastDatabaseRepositories object configured with the default repositories.
        :param database_dir: The root database directory.
        :param is_dist: Whether or not we are building distributable versions of the blast database repositories
            (that is, should we strip out the .git directories).
        :return: The BlastDatabaseRepositories.
        """
        repos = cls(root_database_dir, is_dist)
        repos.register_database_repository('resfinder', 'https://bitbucket.org/genomicepidemiology/resfinder_db.git')
        repos.register_database_repository('pointfinder',
                                           'https://bitbucket.org/genomicepidemiology/pointfinder_db.git')
        repos.register_database_repository('plasmidfinder',
                                           'https://bitbucket.org/genomicepidemiology/plasmidfinder_db.git')

        return repos

    def build_blast_database(self, database_name: str, options: Dict[str, str] = {}) -> AbstractBlastDatabase:
        """
        Builds a staramr.blast.AbstractBlastDatabase from the given parameters.
        :param database_name: The name of the database to build.
        :param options: Options for the particular database in the form of a map {'key': 'value'}
        :return: A new staramr.blast.AbstractBlastDatabase.
        """
        if database_name not in self._database_repositories:
            raise Exception("database_name={} not registered", database_name)

        if database_name == 'resfinder':
            return ResfinderBlastDatabase(self.get_repo_dir(database_name))
        elif database_name == 'pointfinder':
            return PointfinderBlastDatabase(self.get_repo_dir(database_name), options['organism'])
        elif database_name == 'plasmidfinder':
            if options:
                return PlasmidfinderBlastDatabase(self.get_repo_dir(database_name), options['database_type'])
            else:
                return PlasmidfinderBlastDatabase(self.get_repo_dir(database_name))
        else:
            raise Exception("Unknown database name [{}]", database_name)
